/*
 * Renderer 7. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene;

/**
   A {@code Vector} object holds three doubles, which makes it a vector
   in 3-dimensional space.
*/
public final class Vector
{
   public final double x, y, z;

   /**
      Construct a new {@code Vector} using the given x, y, and z coordinates.

      @param x  x-coordinate of the new {@code Vector}
      @param y  y-coordinate of the new {@code Vector}
      @param z  z-coordinate of the new {@code Vector}
   */
   public Vector(final double x, final double y, final double z)
   {
      this.x = x;
      this.y = y;
      this.z = z;
   }


   /**
      A {@code Vector} plus a {@link Vertex} returns a (new) {@link Vertex}.
      <p>
      The vector translates the vertex to a new location.

      @param v  {@link Vertex} object to add to this {@code Vector}
      @return a new {@link Vertex} object that is the translation of {@code v} by this {@code Vector}
   */
   public Vertex plus(final Vertex v)
   {
      return new Vertex(x + v.x, y + v.y, z + v.z);
   }


   /**
      For debugging.

      @return {@link String} representation of this {@code Vector} object
   */
   @Override
   public String toString()
   {
      final int precision = 5;  // the default precision for the format string
      return toString(precision);
   }


   /**
      For debugging.
      <p>
      Allow the precision of the formatted output to be specified.

      @param precision  precision value for the format string
      @return {@link String} representation of this {@code Vector} object
   */
   public String toString(final int precision)
   {
      final int iWidth = 3; // default width of integer part of the format string
      return toString(precision, iWidth);
   }


   /**
      For debugging.
      <p>
      Allow the precision and width of the formatted output to be specified.
      By width, we mean the width of the integer part of each number.

      @param precision  precision value for the format string
      @param iWidth     width of the integer part of the format string
      @return {@link String} representation of this {@code Vector} object
   */
   public String toString(final int precision, final int iWidth)
   {
      // Here is one way to get programmable precision and width.
      final int p = precision;      // the precision for the following format string
      final int t = p + iWidth + 2; // the width for the following format string
      final String format = "[x,y,z] = [% "+t+"."+p+"f  % "+t+"."+p+"f  % "+t+"."+p+"f]";
      return String.format(format, x, y, z);
   }
}
